package CS5350;

import java.io.*;
import java.util.*;

/**
 * Class CMatrix - Class that repr. a Matrix for the purposes of this Proj.
 */
public class CMatrix implements Cloneable {

  // ---------------public Fields-------------------------------
  // Double dim Array with Data [Rows][Cols]
  public int[][] p_iA_Data;

  // ---------------private Fields-------------------------------
  private final static int DEF_SIZE = 10;

  // ------------------------------Constructors-------------------------------

  public CMatrix() {
    p_iA_Data = new int[DEF_SIZE][DEF_SIZE];
  }

  public CMatrix(int i_size) {
    p_iA_Data = new int[i_size][i_size];
  }

  /**
   * CMatrix - Constr. with takes a size and makes it a normal Matr.
   * 
   * @param int -
   *            Matrix size for both rows and cols
   * @param bool-
   *            if given makes it a normal matrix
   */
  public CMatrix(int i_size, boolean b_norm) {
    this(i_size);
    if (b_norm)
      for (int i_i = 0; i_i < i_size; i_i++)
        for (int i_j = 0; i_j < i_size; i_j++)
          p_iA_Data[i_i][i_j] = ((i_i == i_j) ? 1 : 0);
  }

  /**
   * CMatrix - Constr. with takes a size for cols and rows
   * 
   * @param int -
   *            Matrix size for Rows
   * @param int -
   *            Matrix size for Cols
   */
  public CMatrix(int i_rSize, int i_cSize) {
    p_iA_Data = new int[i_rSize][i_cSize];
  }

  /**
   * CMatrix - Constructor out of a file
   * 
   * @param BufferedReader -
   *            BufferedReader over the file
   * @throws IOException
   */
  public CMatrix(BufferedReader BR_theFile) throws IOException {
    String S_read;
    ArrayList<ArrayList> ALAL_rows = new ArrayList<ArrayList>();
    String[] SA_parts;
    int i_colMax = 0;

    // read a line from the file
    while (BR_theFile.ready()) {
      S_read = BR_theFile.readLine();
      if (S_read.startsWith("#"))
        continue;
      // split on whitespace
      SA_parts = S_read.split("\\s");
      // saves max{Cols} 4 later use
      if (SA_parts.length > i_colMax)
        i_colMax = SA_parts.length;
      // make new Row
      ArrayList<Integer> IAL_Columns = new ArrayList<Integer>(i_colMax);
      // and add all cols to it
      for (String S_s : SA_parts)
        IAL_Columns.add(new Integer(S_s));
      // now add to the array of rows
      ALAL_rows.add(IAL_Columns);
    }

    // now we have all the data in a double-dim ArrayList.. move into
    // 'normal' Array
    // and extend shorter columns on the way
    p_iA_Data = new int[ALAL_rows.size()][i_colMax];
    for (int i_i = 0; i_i < ALAL_rows.size(); i_i++) {
      // fill with 0's
      Arrays.fill(p_iA_Data[i_i], 0);

      for (int i_j = 0; i_j < i_colMax; i_j++)
        // and enter into the array only if it was in orig. read line
        // (ie.<>0)
        if (i_j < ALAL_rows.get(i_i).size())
          p_iA_Data[i_i][i_j] = (Integer) ALAL_rows.get(i_i).get(i_j);
    }
  }

  /**
   * CMatrix - Constr. with puts Random stuff in it
   * 
   * @param int -
   *            the symmetric size
   * @param Random -
   *            the RNG to use
   */
  public CMatrix(int i_size, Random R_rng) {
    p_iA_Data = new int[i_size][i_size];
    for (int i_i = 0; i_i < i_size; i_i++)
      for (int i_j = 0; i_j < i_size; i_j++)
        p_iA_Data[i_i][i_j] = R_rng.nextInt();
  }

  /**
   * CMatrix - Constr. with puts Random stuff up to a hardlimit of i_n in it
   * 
   * @param int -
   *            the symmetric size
   * @param Random -
   *            the RNG to use
   */
  public CMatrix(int i_size, Random R_rng, int i_n) {
    p_iA_Data = new int[i_size][i_size];
    for (int i_i = 0; i_i < i_size; i_i++)
      for (int i_j = 0; i_j < i_size; i_j++)
        p_iA_Data[i_i][i_j] = R_rng.nextInt(i_n);
  }

  // ------------------------------Funtions-------------------------------

  /**
   * getRowSize - returns the row size
   * 
   * @return int - the Row size of this instance
   */
  public int getRowSize() {
    return p_iA_Data.length;
  }

  /**
   * getColSize - returns the col size
   * 
   * @return int - the col size of this instance
   */
  public int getColSize() {
    return p_iA_Data[0].length;
  }

  /**
   * invert - returns an inverted copy of this instance, i.e. multiplies all
   * entries by -1
   */
  public CMatrix invert() {
    CMatrix cm = this.clone();
    for (int i_i = 0; i_i < getRowSize(); i_i++)
      for (int i_j = 0; i_j < getColSize(); i_j++)
        cm.p_iA_Data[i_i][i_j] *= -1;
    return cm;
  }

  /**
   * add - adds this instance to another Matrix and returns the outcome
   * 
   * @param CM -
   *            Summand
   * @return CM - the output of the addition as a new Instance
   */
  public CMatrix add(CMatrix CM_sumd) {
    int i_rs = Math.max(getRowSize(), CM_sumd.getRowSize());
    int i_cs = Math.max(getColSize(), CM_sumd.getColSize());
    CMatrix cm = new CMatrix(i_rs, i_cs);

    for (int i_i = 0; i_i < i_rs; i_i++)
      for (int i_j = 0; i_j < i_cs; i_j++) {
        try {
          cm.p_iA_Data[i_i][i_j] = this.p_iA_Data[i_i][i_j]
              + CM_sumd.p_iA_Data[i_i][i_j];
        } catch (Exception e) {
          e.printStackTrace();
        }
      }

    return cm;
  }

  /**
   * sub - subtracts this instance from another Matrix and returns the outcome
   * 
   * @param CM -
   *            Subtractor
   * @return CM - the output of the subtraction as a new Instance
   */
  public CMatrix sub(CMatrix CM_subd) {
    return (this.add(CM_subd.invert()));
  }

  /**
   * multiply - multiplies this instance with another Matrix and returns the
   * outcome
   * 
   * @param CM -
   *            Multiplicator
   * @param IMM -
   *            Interface that implements the mult. function to use
   * @return CM - the output of the multiplication as a new Instance
   */
  public CMatrix multiply(CMatrix CM_cator, IMatrixMulter IMM_func) {
    if (this.getColSize() != CM_cator.getRowSize())
      throw (new RuntimeException("A.cols != B.rows"));
    return IMM_func.MultMatrizes(this, CM_cator);

  }

  /**
   * slice - returns a slice, i.e. a submatrix of this one.
   * 
   * @param int -
   *            row start
   * @param int -
   *            row length
   * @param int -
   *            col start
   * @param int -
   *            col length
   * @return CM - the slice as a new Instance
   */
  public CMatrix slice(int i_rx, int i_rl, int i_cx, int i_cl) {
    CMatrix cm = new CMatrix(i_rl, i_cl);
    for (int i_i = 0; i_i < Math.min(getRowSize() - i_rx, i_rl); i_i++) {
      try {
        System.arraycopy(this.p_iA_Data[i_i + i_rx], // source
            i_cx, // source-index
            cm.p_iA_Data[i_i], // dest
            0, // dest-idx
            Math.min(getColSize() - i_cx, i_cl) // length
            );
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    return cm;
  }

  /**
   * overlay - overlays the given Matrix over this one
   * 
   * @param CM -
   *            the Matrix to overlay
   * @param int -
   *            row start
   * @param int -
   *            col start
   */
  public void overlay(CMatrix CM_set, int i_rx, int i_cx) {
    for (int i_i = 0; i_i < Math.min(getRowSize() - i_rx, CM_set
        .getRowSize()); i_i++) {

      try {
        System.arraycopy(CM_set.p_iA_Data[i_i], 0, this.p_iA_Data[i_i
            + i_rx], i_cx, Math.min(getColSize() - i_cx, CM_set
            .getColSize()));
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  /**
   * toString - return a textual repr. of this Instance
   * 
   * @return String - the text. representation if this Matrix, ie a lot of
   *         numbers
   */
  public String toString() {
    StringBuilder SB_ret = new StringBuilder();

    for (int[] iA_a : p_iA_Data) {
      for (int i_b : iA_a)
        SB_ret.append(i_b + ",");
      SB_ret.append("\r\n");
    }

    return new String("Matrix{\r\n" + SB_ret + "}\r\n");
  }

  /**
   * clone - returns a deep copy of this instance
   * 
   * @return CM - the copy
   */
  public CMatrix clone() {
    CMatrix cm = new CMatrix(getRowSize(), getColSize());
    for (int i_i = 0; i_i < getRowSize(); i_i++)
      System.arraycopy(p_iA_Data[i_i], 0, cm.p_iA_Data[i_i], 0,
          getColSize());
    return cm;
  }

  /**
   * equals - decide wether two instances of CM are equal
   * 
   * @param Object -
   *            the Object to compare with
   * @return bool - (this == o)
   */
  public boolean equals(Object O_o) {
    if (!(O_o instanceof CMatrix))
      return false;

    boolean ret = true;
    for (int i_i = 0; i_i < getRowSize(); i_i++)
      for (int i_j = 0; i_j < getColSize(); i_j++) {
        try {
          ret = (p_iA_Data[i_i][i_j] == ((CMatrix) O_o).p_iA_Data[i_i][i_j]);
        } catch (Exception e) {
          return false;
        }
      }
    return ret;
  }

}
