package CS5350;

import java.util.*;
import java.lang.management.*;

/**
 * @author Vitus
 * @version 0.2
 * @written 1.10.2k5;11:15
 */

/**
 * Class CMatrixMult_Main - main class for the Prog
 */
public class CMatrixMult_Main {

  // some constants - coz 'magic' numbers are just wrong
  private static int POW_BASE = 10;

  private static int STRASSEN_SIZE = 8;

  // number of distinct points we can get out of Strassens
  private static int DIST_POINTS = (int) Math.ceil(Math.log(STRASSEN_SIZE)
      / Math.log(2));

  private static int ITERATIONS = 5000;

  private static int ROUND_TO = (int) Math.pow(POW_BASE, 3);

  private static int US_DIV = (int) Math.pow(POW_BASE, 3);

  private static int MS_DIV = (int) Math.pow(POW_BASE, 6);

  private static int S_DIV = (int) Math.pow(POW_BASE, 9);

  // order of the poly we are trying to fit
  private static int ORDER_POLY = 2;

  private static int RES_DELTA_TIMES = 50;

  // monitoring Bean
  private static ThreadMXBean s_TMXB_Tmxb;

  // array of the x'es
  private double[] dA_Xs = new double[DIST_POINTS];

  // array of the y'es
  private double[] dA_Ys = new double[DIST_POINTS];

  // ------------------------------Constructors-------------------------------

  /**
   * CMatrixMult_Main - empty Constructor
   */
  CMatrixMult_Main() {
    // try resolution of nano
    long l_start = System.nanoTime(), l_cur = 0, l_last = l_start, l_delta = l_start;
    for (int i_i = 0; i_i < RES_DELTA_TIMES; i_i++) {
      l_cur = System.nanoTime();
      if ((l_cur - l_last) < l_delta)
        l_delta = (l_cur - l_last);
      l_last = l_cur;
    }
    System.out.println("TimerRes=" + l_delta + "ns");
    // try to get ThreadManMXBean 4 Thread Time Monitoring
    s_TMXB_Tmxb = ManagementFactory.getThreadMXBean();
    s_TMXB_Tmxb.setThreadCpuTimeEnabled(true);

    // create matrizes
    CMatrix a = new CMatrix(STRASSEN_SIZE, new Random());
    CMatrix b = new CMatrix(STRASSEN_SIZE, new Random());
    // and multiply using naive
    CMatrix c = callPart1(a, b);
    // multiply using strassen
    CMatrix d = callPart2(a, b);
    // are they equal?
    System.out.println("(naive=strassen)?" + c.equals(d));

    // now try to get poly Least Squared Fit on x's and y's using no
    // error
    double[] dA_f = new double[(ORDER_POLY + 1) * 2];
    float[] fA_f = new float[(ORDER_POLY + 1) * 2];
    // fit the poly, gives the coeff back in dA_f[0..Order+1/2] and
    // Errors in dA_f[Order+1/2..Order+1]
    FitPoly.fit(dA_f, dA_Xs, dA_Ys, null, null, (ORDER_POLY + 1));

    // norm highest coeff to 1
    for (int i_i = 0; i_i < (ORDER_POLY + 1); i_i++)
      fA_f[i_i] = ((float) dA_f[i_i] / (float) dA_f[ORDER_POLY]);

    // round the numbers off
    for (int i_i = 0; i_i < (ORDER_POLY + 1); i_i++) {
      // Poly coeffs
      fA_f[i_i] = ((float) Math.round((float) ROUND_TO * fA_f[i_i]))
          / ROUND_TO;
      // Error coeffs
      fA_f[i_i + (ORDER_POLY + 1)] = ((float) Math.round((float) ROUND_TO
          * dA_f[i_i + (ORDER_POLY + 1)]))
          / ROUND_TO;
    }
    // and print the coeffs with acc. errors
    System.out.println("------------------");

    System.out.print("Poly= x^" + ORDER_POLY + " + ");
    // print out Poly coeffs depending on how many there are
    for (int i_i = ORDER_POLY - 1; i_i >= 0; i_i--) {
      String S_tail = "";
      if (i_i == 1)
        S_tail = "x";
      else if (i_i > 1)
        S_tail = ("x^" + i_i);
      System.out.print("(" + fA_f[i_i] + ")" + S_tail);
      if (i_i > 0)
        System.out.print(" + ");
    }
    System.out.println();
    // print out Error coeffs depending on how many there are
    System.out.print("ErrP= ");
    for (int i_i = ORDER_POLY; i_i >= 0; i_i--) {
      String S_tail = "";
      if (i_i == 1)
        S_tail = "x";
      else if (i_i > 1)
        S_tail = ("x^" + i_i);
      System.out.print("(" + fA_f[i_i + (ORDER_POLY + 1)] + ")" + S_tail);
      if (i_i > 0)
        System.out.print(" + ");
    }
    System.out.println();

    System.out.println("------------------");
    // print the whole time passed
    System.out.println("sysTime passed=" + (System.nanoTime() - l_start)
        / S_DIV + "s");
    // and finally print how many time this thread took (not very useful)
    System.out.println("mainThread UserTime="
        + (s_TMXB_Tmxb.getCurrentThreadUserTime() / MS_DIV) + "ms");

  }

  /**
   * callPart1 - func that calls the naive implementation
   * 
   * @param CM -
   *            Multiplicant
   * @param CM -
   *            Multiplicator
   */
  private CMatrix callPart1(CMatrix CM_a, CMatrix CM_b) {

    // test all Matrix sizes that strassen also uses to get some sense of
    // the overhead
    for (int i_i = 1; i_i < (Math.log(STRASSEN_SIZE) / Math.log(2)); i_i++) {

      CMatrix a = new CMatrix((int) Math.pow(2, i_i), new Random());
      CMatrix b = new CMatrix((int) Math.pow(2, i_i), new Random());
      CRunner CR_cRunner = new CRunner(a, b, ITERATIONS
          * Math.pow(10, 5 - i_i));
      CR_cRunner.start();
      try {
        CR_cRunner.join();
      } catch (Exception e) {
        System.out.println("Exception in join:" + e);
      }
      System.out.println("naive("
          + a.getRowSize()
          + ")TotalTime="
          + formatHug(CR_cRunner.p_d_Spent, ITERATIONS
              * Math.pow(10, 5 - i_i)));
    }
    // finally compute the final answer
    CRunner CR_cRunner = new CRunner(CM_a, CM_b, ITERATIONS);
    CR_cRunner.start();
    try {
      CR_cRunner.join();
    } catch (Exception e) {
      System.out.println("Exception in join:" + e);
    }
    // print the total used time for naive
    System.out.println("naive(" + CM_a.getRowSize() + ")TotalTime="
        + formatHug(CR_cRunner.p_d_Spent, ITERATIONS));
    // and return
    return CR_cRunner.CM_C;
  }

  /**
   * callPart2 - func that calls Strassens algo
   * 
   * @param CM -
   *            Multiplicant
   * @param CM -
   *            Multiplicator
   */
  private CMatrix callPart2(CMatrix CM_a, CMatrix CM_b) {

    // kick of Strassen in its own thread
    CPart2 CP_cp2 = new CPart2(CM_a, CM_b, ITERATIONS);
    Thread T_t = new Thread(CP_cp2);
    T_t.start();
    try {
      T_t.join();
    } catch (Exception e) {
      e.printStackTrace();
    }

    int i_j = -1;
    // check the array of stored times for values that exist and print them
    for (int i_i = 0; i_i < CP_cp2.ps_dA_Overheads.length; i_i++) {
      if (CP_cp2.ps_dA_Overheads[i_i] != 0) {
        System.out
            .println("strassen("
                + (i_i + 1)
                + ")Overhead="
                + formatHug(
                    (CP_cp2.ps_dA_Overheads[i_i] / CP_cp2.ps_dA_Overheads_times[i_i]),
                    ITERATIONS));
        // and store in array for least square fit
        if ((++i_j) < DIST_POINTS) {
          dA_Xs[i_j] = (i_i + 1);
          dA_Ys[i_j] = ((float) CP_cp2.ps_dA_Overheads[i_i]
              / CP_cp2.ps_dA_Overheads_times[i_i] / ITERATIONS);
        }
      }
    }
    // print the total used time
    System.out.println("strassen(" + CM_a.getRowSize() + ")TotalTime="
        + formatHug(CP_cp2.ps_d_TotalExecTime, ITERATIONS));

    return CP_cp2.getRes();

  }

  /**
   * Class CRunner - private class that represents the thread which does the
   * work
   */
  private class CRunner extends Thread {
    public CMatrix CM_A, CM_B, CM_C;

    private IMatrixMulter IMM_Imm = new CPart1();

    private double d_Times = 1; // times to iterate

    public double p_d_Spent; // single iteration time used / ns

    /**
     * CRunner - Constr. with the two Matr. to mult.
     * 
     * @param CM -
     *            Multiplicant
     * @param CM -
     *            Multiplicator
     */
    CRunner(CMatrix CM_a, CMatrix CM_b) {
      CM_A = CM_a;
      CM_B = CM_b;
      // set prio to high
      setPriority(MAX_PRIORITY);
    }

    /**
     * CRunner - Constr. with the two Matr. to mult.
     * 
     * @param CM -
     *            Multiplicant
     * @param CM -
     *            Multiplicator
     * @param d_t -
     *            Times to iterate
     */
    CRunner(CMatrix CM_a, CMatrix CM_b, double d_t) {
      this(CM_a, CM_b);
      d_Times = d_t;
    }

    /**
     * run - method that will be run inside this thread object does the
     * actual work of multiplicating
     */
    public void run() {

      // and run the mult'ing X specified i_Times
      for (int i_i = 0; i_i < d_Times; i_i++)
        CM_C = CM_A.multiply(CM_B, IMM_Imm);

      // get time spent User
      p_d_Spent = s_TMXB_Tmxb.getCurrentThreadUserTime();
      
    }

  }

  /**
   * formatHug - little helper function that takes a time and iterations and
   * formats it in one of {ns,us,ms,s}
   * 
   * @param args -
   *            command line args
   */
  public static String formatHug(double d_spent, double d_times) {
    String S_single = "";

    // give the output a nice little format hug *hugs*
    if ((d_spent / d_times) < US_DIV) {
      S_single = d_spent / d_times + "ns";
    } else if ((d_spent / d_times) < MS_DIV) {
      S_single = ((d_spent / d_times) / US_DIV) + "us";
    } else if ((d_spent / d_times) < S_DIV) {
      S_single = ((d_spent / d_times) / MS_DIV) + "ms";
    } else if ((d_spent / d_times) < Math.pow(10, 12)) {
      S_single = ((d_spent / d_times) / S_DIV) + "s";
    }

    return S_single;
  }

  /**
   * main - main program
   * 
   * @param args -
   *            command line args
   */
  public static void main(String[] args) {
    new CMatrixMult_Main();

  }

}
