package CS5350;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;

/**
 * Class CPart2 - Class that implements a Matrix mult. using Strassen's Algo
 * completes the second part of Project assignment
 */
public class CPart2 implements IMatrixMulter, Runnable {

  private static ThreadMXBean s_TMXB_Tmxb;

  private static CMatrix s_CM_A;

  private static CMatrix s_CM_B;

  private static CMatrix s_CM_C;

  // public's
  public static int ps_i_It = 1000;

  public static double[] ps_dA_Overheads;

  public static int[] ps_dA_Overheads_times;

  public static double ps_d_TotalExecTime;

  /**
   * Constructor with two matrizes
   * 
   * @param CM -
   *            first Matrix
   * @param CM -
   *            sec. Matrix
   */
  public CPart2(CMatrix CM_a, CMatrix CM_b, int i_it) {
    s_CM_A = CM_a;
    s_CM_B = CM_b;
    ps_i_It = i_it;
    ps_dA_Overheads = new double[CM_a.getRowSize()];
    ps_dA_Overheads_times = new int[CM_a.getRowSize()];
  }

  /**
   * run - overrides the abstract run in runnable used to run this class in
   * stand-alone-mode
   */
  public void run() {
    s_CM_C = this.MultMatrizes(s_CM_A, s_CM_B);
    ps_d_TotalExecTime = s_TMXB_Tmxb.getCurrentThreadUserTime();
  }

  /**
   * getRes - returns the result of a run used for the stand-alone-mode
   */
  public CMatrix getRes() {
    return this.s_CM_C;
  }

  /**
   * MultMatrizes - implements matrix mult using strassen's algo
   * 
   * @param CM -
   *            first Matrix (Multiplicant)
   * @param CM -
   *            sec. Matrix (Multiplicator)
   * @return CM - the product of a and b
   */
  public CMatrix MultMatrizes(CMatrix CM_a, CMatrix CM_b) {
    int i_n = CM_a.getRowSize();
    if ((i_n != CM_a.getColSize()) || (i_n != CM_b.getRowSize())
        || (i_n != CM_b.getColSize()))
      throw (new RuntimeException("not (n x n)(n x n)"));

    s_TMXB_Tmxb = ManagementFactory.getThreadMXBean();
    s_TMXB_Tmxb.setThreadCpuTimeEnabled(true);

    return strassen(CM_a, CM_b, i_n);

  }

  /**
   * strassen - implements matrix mult using strassen's algo
   *        this is the recursive function
   * @param CM -
   *            first Matrix (Multiplicant)
   * @param CM -
   *            sec. Matrix (Multiplicator)
   * @param int -
   *            the (quadr.) size of a and b
   * @return CM - the product of a and b
   */
  private CMatrix strassen(CMatrix CM_a, CMatrix CM_b, int i_n) {
    double d_spent = 0;
    double d_start = 0;
    CMatrix CM_return = new CMatrix(i_n, i_n);

    // if matrix size is 1 dont start strassen, just multiplicate and return
    if (i_n == 1) {
      CM_return.p_iA_Data[0][0] = CM_a.p_iA_Data[0][0]
          * CM_b.p_iA_Data[0][0];
    } else {

      // half the size, rounded up
      int i_h = (int) Math.ceil(((float) i_n) / 2);

      /** measuring of time starts here: save first checkpoint */
      d_start = s_TMXB_Tmxb.getCurrentThreadUserTime();

      CMatrix a = null, b = null, c = null, d = null, e = null, f = null;
      CMatrix g = null, h = null, tp1 = null, tp2 = null, tp3 = null;
      CMatrix tp4 = null, tp5a = null, tp5b = null, tp6a = null;
      CMatrix tp6b = null, tp7a = null, tp7b = null;

      // start iteration to get better timing results
      for (int i_i = 0; i_i < ps_i_It; i_i++) {

        // compute the 8 slices row_s row_l col_s col_l
        a = CM_a.slice(0, i_h, 0, i_h);
        b = CM_a.slice(0, i_h, i_h, i_h);
        c = CM_a.slice(i_h, i_h, 0, i_h);
        d = CM_a.slice(i_h, i_h, i_h, i_h);

        e = CM_b.slice(0, i_h, 0, i_h);
        f = CM_b.slice(0, i_h, i_h, i_h);
        g = CM_b.slice(i_h, i_h, 0, i_h);
        h = CM_b.slice(i_h, i_h, i_h, i_h);

        // compute temporary variables for P1..P7
        tp1 = f.sub(h);
        tp2 = a.add(b);
        tp3 = c.add(d);
        tp4 = g.sub(e);
        tp5a = a.add(d);
        tp5b = e.add(h);
        tp6a = b.sub(d);
        tp6b = g.add(h);
        tp7a = a.sub(c);
        tp7b = e.add(f);
      }

      /** stop the time measuring by saving an intermediate value */
      d_spent = s_TMXB_Tmxb.getCurrentThreadUserTime();

      /** compute real P1..P7 - recursive part */
      CMatrix P1 = strassen(a, tp1, i_h);
      CMatrix P2 = strassen(tp2, h, i_h);
      CMatrix P3 = strassen(tp3, e, i_h);
      CMatrix P4 = strassen(d, tp4, i_h);
      CMatrix P5 = strassen(tp5a, tp5b, i_h);
      CMatrix P6 = strassen(tp6a, tp6b, i_h);
      CMatrix P7 = strassen(tp7a, tp7b, i_h);
      // System.out.println("----------------");
      /** end recursive part */

      /** start it again by computing the time that was spent in rec. part */
      d_spent = s_TMXB_Tmxb.getCurrentThreadUserTime() - d_spent;

      for (int i_i = 0; i_i < ps_i_It; i_i++) {

        // now combine them to make the four parts of result
        CMatrix r = P5.add(P4).sub(P2).add(P6);
        CMatrix s = P1.add(P2);
        CMatrix t = P3.add(P4);
        CMatrix u = P5.add(P1).sub(P3).sub(P7);

        // and overlay them over the result matrix
        CM_return.overlay(r, 0, 0);
        CM_return.overlay(s, 0, i_h);
        CM_return.overlay(t, i_h, 0);
        CM_return.overlay(u, i_h, i_h);
      }

      /** and finally get full time and sub time spent in rec. part */
      d_spent = s_TMXB_Tmxb.getCurrentThreadUserTime() - d_start
          - d_spent;
      
//      store used times for later lsf
      ps_dA_Overheads[i_n - 1] += d_spent;
      ps_dA_Overheads_times[i_n - 1]++;
    }

    return CM_return;
  }
}
